/* tarot implements the rules of the tarot game
 * Copyright (C) 2019  Vivien Kraus
 *
 * This program is free software: you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation, version 3 of the License.
 *
 * This program is distributed in the hope that it will be useful, but
 * WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program.  If not, see <http://www.gnu.org/licenses/>.
 */

#ifndef H_TAROT_FEATURES_PRIVATE_IMPL_INCLUDED
#define H_TAROT_FEATURES_PRIVATE_IMPL_INCLUDED

#include <tarot/game_private.h>
#include <tarot/game_event_private.h>
#include <tarot/cards_private.h>
#include <tarot/counter_private.h>
#include <tarot/cnn_private.h>
#include <stdio.h>
#include <stdlib.h>
#include <assert.h>
#include <string.h>
#include <float.h>
#include <errno.h>
#include "xalloc.h"

static char *features_header_data = NULL;
static char *features_header_encoded_data = NULL;

static inline FILE *
features_fopen_in_dir (const char *dir, const char *basename)
{
  FILE *f;
  char *filename =
    xmalloc (strlen (dir) + strlen ("/") + strlen (basename) + 1);
  strcpy (filename, dir);
  strcat (filename, "/");
  strcat (filename, basename);
  f = fopen (filename, "r");
  if (f == NULL)
    {
      fprintf (stderr, _("Error: could not open file %s: %s\n"), filename,
               strerror (errno));
      abort ();
    }
  free (filename);
  return f;
}

static inline void
features_buffer_push (size_t *allocated, char **buffer, size_t *used, char c)
{
  if (*allocated == *used)
    {
      *allocated *= 2;
      *buffer = xrealloc (*buffer, *allocated);
    }
  assert (*used < *allocated);
  (*buffer)[(*used)++] = c;
}

static inline char *
features_read_whole_file (FILE * f)
{
  size_t allocated = 1;
  char *buffer = xmalloc (allocated);
  size_t used = 0;
  int i;
  while (1)
    {
      i = fgetc (f);
      if (i == EOF)
        {
          features_buffer_push (&allocated, &buffer, &used, 0);
          return buffer;
        }
      else
        {
          features_buffer_push (&allocated, &buffer, &used,
                                (char) (unsigned char) (unsigned int) i);
        }
    }
}

static inline const char *
features_arff_header (const char *pkgdatadir, int encoded)
{
  if (encoded && features_header_encoded_data == NULL)
    {
      FILE *f =
        features_fopen_in_dir (pkgdatadir, "features-header-encoded.txt");
      static const char *sought = "@relation tarot-" VERSION "-encoded";
      features_header_encoded_data = features_read_whole_file (f);
      fclose (f);
      if (strncmp (features_header_encoded_data, sought, strlen (sought)) !=
          0)
        {
          fprintf (stderr,
                   _
                   ("Error: the installed ARFF header (%s/features-header-encoded.txt) is not compatible with this version (%s).\n"),
                   pkgdatadir, VERSION);
          abort ();
        }
      return features_header_encoded_data;
    }
  else if (encoded)
    {
      return features_header_encoded_data;
    }
  else if (!encoded && features_header_data == NULL)
    {
      FILE *f = features_fopen_in_dir (pkgdatadir, "features-header.txt");
      static const char *sought = "@relation tarot-" VERSION;
      features_header_data = features_read_whole_file (f);
      fclose (f);
      if (strncmp (features_header_data, sought, strlen (sought)) != 0)
        {
          fprintf (stderr,
                   _
                   ("Error: the installed ARFF header (%s/features-header.txt) is not compatible with this version (%s).\n"),
                   pkgdatadir, VERSION);
          abort ();
        }
      return features_header_data;
    }
  else if (!encoded)
    {
      return features_header_data;
    }
  assert (0);
  return NULL;
}

static size_t
features_data_push (char ch, size_t *max, char **dest)
{
  if (*max == 1)
    {
      **dest = '\0';
      *max = 0;
      *dest = NULL;
    }
  else if (*max != 0)
    {
      **dest = ch;
      *dest = *dest + 1;
      *max = *max - 1;
    }
  return 1;
}

static size_t
features_data_push_string (const char *str, size_t *max, char **dest)
{
  size_t ret = 0;
  while (*str != '\0')
    {
      ret += features_data_push (*str, max, dest);
      str++;
    }
  return ret;
}

static size_t
features_data_push_int (int i, size_t *max, char **dest)
{
  char repr[32];
  snprintf (repr, sizeof (repr), "%d", i);
  return features_data_push_string (repr, max, dest);
}

static size_t
features_data_push_ordered (int mini, int maxi, int value, size_t *max,
                            char **dest)
{
  static const char *sep[] = {
    "",
    ","
  };
  int has_sep = 0;
  int i;
  size_t ret = 0;
  assert (value >= mini);
  assert (value <= maxi);
  for (i = mini; i <= value; i++)
    {
      ret += features_data_push_string (sep[has_sep], max, dest);
      ret += features_data_push_string ("1", max, dest);
      has_sep = 1;
    }
  for (i = value + 1; i <= maxi; i++)
    {
      ret += features_data_push_string (sep[has_sep], max, dest);
      ret += features_data_push_string ("0", max, dest);
      has_sep = 1;
    }
  return ret;
}

static size_t
features_data_push_yesno (int v, size_t *max, char **dest)
{
  if (v)
    {
      return features_data_push_string ("yes", max, dest);
    }
  return features_data_push_string ("no", max, dest);
}

#define INT(i) ret += features_data_push_int ((i), &max, &dest);
#define STRING(s) ret += features_data_push_string (s, &max, &dest);
#define COMMA STRING (",");
#define ORDERED(mini, maxi, value) ret += features_data_push_ordered (mini, maxi, value, &max, &dest);
#define YESNO(v) ret += features_data_push_yesno (v, &max, &dest);

static inline size_t
features_arff_data (const TarotGame * game, const TarotGameEvent * added,
                    int score, int encoded, size_t max, char *dest)
{
  size_t f_n_players;
  int f_with_call;
  size_t f_trick;
  size_t f_position;
  int f_partner_known;
  TarotBid f_bid_minimum;
  size_t f_n_kings;
  size_t f_n_marriages;
  size_t f_n_extended_marriages;
  size_t f_n_queens;
  size_t f_n_knights;
  size_t f_n_shortest;
  size_t f_n_second_shortest;
  size_t f_n_second_longest;
  size_t f_n_longest;
  int f_has_petit;
  int f_has_excuse;
  int f_has_twentyone;
  size_t f_n_trumps;
  TarotNumber f_max_trump;
  TarotNumber f_min_trump;
  TarotNumber f_q1_trumps;
  TarotNumber f_q3_trumps;
  TarotNumber f_median_trumps;
  int f_team_attack;
  int f_team_defence;
  size_t f_n_trick_players_already_played;
  size_t f_n_allies_after_last_opponent;
  int f_play_called_suit;
  size_t f_play_n_times_lead;
  size_t f_play_suit_remaining;
  size_t f_play_suit_remaining_points;
  int f_ally_may_be_minor_master;
  int f_ennemy_may_be_minor_master;
  size_t f_n_trumps_remaining;
  int f_ally_may_be_trump_master;
  int f_ennemy_may_be_trump_master;
  size_t f_n_halfpoints_in_trick;
  size_t f_n_oudlers_in_trick;
  TarotBid s_bid;
  int s_outbid;
  int s_call_shortest;
  int s_call_second_shortest;
  int s_call_second_longest;
  int s_call_longest;
  int s_call_myself;
  int s_call_marriage;
  size_t s_discard_queens;
  size_t s_discard_knights;
  size_t s_discard_jacks;
  size_t s_discard_junk;
  size_t s_discard_shortest;
  size_t s_discard_second_shortest;
  size_t s_discard_second_longest;
  size_t s_discard_longest;
  int s_lead_called_suit;
  size_t s_lead_suit_length;
  size_t s_lead_n_times_lead;
  int s_lead_trumps;
  size_t s_n_smaller_kept;
  size_t s_n_greater_nonface_kept;
  size_t s_n_points_kept;
  size_t s_n_greater_kept;
  int s_play_excuse;
  int s_trump_master;
  int s_trump_obvious_master;
  size_t s_halfpoints;
  int s_oudler;
  TarotSuit order[4];
  TarotCounter counter;
  size_t ret = 0;
  counter_initialize (&counter);
  counter_load_from_game (&counter, game);
  features_variant (game, &f_n_players, &f_with_call);
  features_position (game, &f_trick, &f_position);
  features_partner_known (game, &f_partner_known);
  features_bid_minimum (game, &f_bid_minimum);
  features_count_faces (game, &f_n_kings, &f_n_marriages,
                        &f_n_extended_marriages, &f_n_queens, &f_n_knights);
  features_count_minor (game, &f_n_shortest, &f_n_second_shortest,
                        &f_n_second_longest, &f_n_longest, order);
  features_count_oudlers (game, &f_has_petit, &f_has_excuse,
                          &f_has_twentyone);
  features_count_trumps (game, &f_n_trumps, &f_max_trump, &f_min_trump,
                         &f_q1_trumps, &f_q3_trumps, &f_median_trumps);
  features_team (game, &f_team_attack, &f_team_defence);
  features_trick_players (game, &f_n_trick_players_already_played,
                          &f_n_allies_after_last_opponent);
  features_minor_suit (game, &counter, &f_play_called_suit,
                       &f_play_n_times_lead, &f_play_suit_remaining,
                       &f_play_suit_remaining_points,
                       &f_ally_may_be_minor_master,
                       &f_ennemy_may_be_minor_master);
  features_trump (game, &counter, &f_n_trumps_remaining,
                  &f_ally_may_be_trump_master, &f_ennemy_may_be_trump_master);
  features_points (game, &f_n_halfpoints_in_trick, &f_n_oudlers_in_trick);
  strategy_bid (game, added, &s_bid);
  strategy_outbid (game, added, &s_outbid);
  strategy_call_by_length (game, added, &s_call_shortest,
                           &s_call_second_shortest,
                           &s_call_second_longest, &s_call_longest);
  strategy_call_myself (game, added, &s_call_myself);
  strategy_call_marriage (game, added, &s_call_marriage);
  strategy_discard_save (game, added, &s_discard_queens,
                         &s_discard_knights, &s_discard_jacks);
  strategy_discard_junk (game, added, &s_discard_junk);
  strategy_discard_by_length (game, added, &s_discard_shortest,
                              &s_discard_second_shortest,
                              &s_discard_second_longest, &s_discard_longest);
  strategy_lead (game, &counter, added, &s_lead_called_suit,
                 &s_lead_suit_length, &s_lead_n_times_lead, &s_lead_trumps);
  strategy_purity (game, added, &s_n_smaller_kept, &s_n_greater_nonface_kept,
                   &s_n_points_kept, &s_n_greater_kept);
  strategy_excuse (game, added, &s_play_excuse);
  strategy_trump_master (game, &counter, added, &s_trump_master,
                         &s_trump_obvious_master);
  strategy_play_points (game, added, &s_halfpoints, &s_oudler);
  if (encoded)
    {
      INT (f_n_players == 3);
      COMMA;
      INT (f_n_players == 4);
      COMMA;
      INT (f_n_players == 5);
      COMMA;
      INT (f_with_call);
      COMMA;
      ORDERED (0, 24, f_trick);
      COMMA;
      ORDERED (0, 5, f_position);
      COMMA;
      INT (tarot_game_event_type (added) == TAROT_BID_EVENT);
      COMMA;
      INT (tarot_game_event_type (added) == TAROT_DECL_EVENT);
      COMMA;
      INT (tarot_game_event_type (added) == TAROT_CALL_EVENT);
      COMMA;
      INT (tarot_game_event_type (added) == TAROT_DISCARD_EVENT);
      COMMA;
      INT (tarot_game_event_type (added) == TAROT_HANDFUL_EVENT);
      COMMA;
      INT (tarot_game_event_type (added) == TAROT_CARD_EVENT);
      COMMA;
      INT (f_partner_known);
      COMMA;
      INT (f_bid_minimum == TAROT_PASS);
      COMMA;
      INT (f_bid_minimum == TAROT_TAKE);
      COMMA;
      INT (f_bid_minimum == TAROT_PUSH);
      COMMA;
      INT (f_bid_minimum == TAROT_STRAIGHT_KEEP);
      COMMA;
      INT (f_bid_minimum == TAROT_DOUBLE_KEEP);
      COMMA;
      ORDERED (0, 4, f_n_kings);
      COMMA;
      ORDERED (0, 4, f_n_marriages);
      COMMA;
      ORDERED (0, 4, f_n_extended_marriages);
      COMMA;
      ORDERED (0, 4, f_n_queens);
      COMMA;
      ORDERED (0, 4, f_n_knights);
      COMMA;
      ORDERED (0, 14, f_n_shortest);
      COMMA;
      ORDERED (0, 14, f_n_second_shortest);
      COMMA;
      ORDERED (0, 14, f_n_second_longest);
      COMMA;
      ORDERED (0, 14, f_n_longest);
      COMMA;
      INT (f_has_petit);
      COMMA;
      INT (f_has_excuse);
      COMMA;
      INT (f_has_twentyone);
      COMMA;
      ORDERED (0, 21, f_n_trumps);
      COMMA;
      ORDERED (0, 21, f_max_trump);
      COMMA;
      ORDERED (0, 21, f_min_trump);
      COMMA;
      ORDERED (0, 21, f_q1_trumps);
      COMMA;
      ORDERED (0, 21, f_q3_trumps);
      COMMA;
      ORDERED (0, 21, f_median_trumps);
      COMMA;
      INT (f_team_attack);
      COMMA;
      INT (f_team_defence);
      COMMA;
      ORDERED (0, 4, f_n_trick_players_already_played);
      COMMA;
      ORDERED (0, 5, f_n_allies_after_last_opponent);
      COMMA;
      INT (f_play_called_suit);
      COMMA;
      ORDERED (0, 21, f_play_n_times_lead);
      COMMA;
      ORDERED (0, 21, f_play_suit_remaining);
      COMMA;
      ORDERED (0, 37, f_play_suit_remaining_points);
      COMMA;
      INT (f_ally_may_be_minor_master);
      COMMA;
      INT (f_ennemy_may_be_minor_master);
      COMMA;
      ORDERED (0, 21, f_n_trumps_remaining);
      COMMA;
      INT (f_ally_may_be_trump_master);
      COMMA;
      INT (f_ennemy_may_be_trump_master);
      COMMA;
      ORDERED (0, 45, f_n_halfpoints_in_trick);
      COMMA;
      ORDERED (0, 3, f_n_oudlers_in_trick);
      COMMA;
      INT (s_bid == TAROT_PASS);
      COMMA;
      INT (s_bid == TAROT_TAKE);
      COMMA;
      INT (s_bid == TAROT_PUSH);
      COMMA;
      INT (s_bid == TAROT_STRAIGHT_KEEP);
      COMMA;
      INT (s_bid == TAROT_DOUBLE_KEEP);
      COMMA;
      INT (s_outbid);
      COMMA;
      INT (s_call_shortest);
      COMMA;
      INT (s_call_second_shortest);
      COMMA;
      INT (s_call_second_longest);
      COMMA;
      INT (s_call_longest);
      COMMA;
      INT (s_call_myself);
      COMMA;
      INT (s_call_marriage);
      COMMA;
      ORDERED (0, 4, s_discard_queens);
      COMMA;
      ORDERED (0, 4, s_discard_knights);
      COMMA;
      ORDERED (0, 4, s_discard_jacks);
      COMMA;
      ORDERED (0, 40, s_discard_junk);
      COMMA;
      ORDERED (0, 6, s_discard_shortest);
      COMMA;
      ORDERED (0, 6, s_discard_second_shortest);
      COMMA;
      ORDERED (0, 6, s_discard_second_longest);
      COMMA;
      ORDERED (0, 6, s_discard_longest);
      COMMA;
      INT (s_lead_called_suit);
      COMMA;
      ORDERED (0, 21, s_lead_suit_length);
      COMMA;
      ORDERED (0, 21, s_lead_n_times_lead);
      COMMA;
      INT (s_lead_trumps);
      COMMA;
      ORDERED (0, 21, s_n_smaller_kept);
      COMMA;
      ORDERED (0, 21, s_n_greater_nonface_kept);
      COMMA;
      ORDERED (0, 37, s_n_points_kept);
      COMMA;
      ORDERED (0, 21, s_n_greater_kept);
      COMMA;
      INT (s_play_excuse);
      COMMA;
      INT (s_trump_master);
      COMMA;
      INT (s_trump_obvious_master);
      COMMA;
      ORDERED (0, 9, s_halfpoints);
      COMMA;
      INT (s_oudler);
    }
  else
    {
      INT ((int) f_n_players);
      COMMA;
      YESNO (f_with_call);
      COMMA;
      INT ((int) f_trick);
      COMMA;
      INT ((int) f_position);
      switch (tarot_game_event_type (added))
        {
        case TAROT_BID_EVENT:
          COMMA;
          STRING ("TAROT_BID_EVENT");
          break;
        case TAROT_DECL_EVENT:
          COMMA;
          STRING ("TAROT_DECL_EVENT");
          break;
        case TAROT_CALL_EVENT:
          COMMA;
          STRING ("TAROT_CALL_EVENT");
          break;
        case TAROT_DISCARD_EVENT:
          COMMA;
          STRING ("TAROT_DISCARD_EVENT");
          break;
        case TAROT_HANDFUL_EVENT:
          COMMA;
          STRING ("TAROT_HANDFUL_EVENT");
          break;
        case TAROT_CARD_EVENT:
          COMMA;
          STRING ("TAROT_CARD_EVENT");
          break;
        default:
          assert (0);
        }
      COMMA;
      YESNO (f_partner_known);
      switch (f_bid_minimum)
        {
        case TAROT_PASS:
          COMMA;
          STRING ("TAROT_PASS");
          break;
        case TAROT_TAKE:
          COMMA;
          STRING ("TAROT_TAKE");
          break;
        case TAROT_PUSH:
          COMMA;
          STRING ("TAROT_PUSH");
          break;
        case TAROT_STRAIGHT_KEEP:
          COMMA;
          STRING ("TAROT_STRAIGHT_KEEP");
          break;
        case TAROT_DOUBLE_KEEP:
          COMMA;
          STRING ("TAROT_DOUBLE_KEEP");
          break;
        default:
          assert (0);
        }
      COMMA;
      INT ((int) f_n_kings);
      COMMA;
      INT ((int) f_n_marriages);
      COMMA;
      INT ((int) f_n_extended_marriages);
      COMMA;
      INT ((int) f_n_queens);
      COMMA;
      INT ((int) f_n_knights);
      COMMA;
      INT ((int) f_n_shortest);
      COMMA;
      INT ((int) f_n_second_shortest);
      COMMA;
      INT ((int) f_n_second_longest);
      COMMA;
      INT ((int) f_n_longest);
      COMMA;
      YESNO (f_has_petit);
      COMMA;
      YESNO (f_has_excuse);
      COMMA;
      YESNO (f_has_twentyone);
      COMMA;
      INT ((int) f_n_trumps);
      COMMA;
      INT ((int) f_max_trump);
      COMMA;
      INT ((int) f_min_trump);
      COMMA;
      INT ((int) f_q1_trumps);
      COMMA;
      INT ((int) f_q3_trumps);
      COMMA;
      INT ((int) f_median_trumps);
      COMMA;
      YESNO (f_team_attack);
      COMMA;
      YESNO (f_team_defence);
      COMMA;
      INT ((int) f_n_trick_players_already_played);
      COMMA;
      INT ((int) f_n_allies_after_last_opponent);
      COMMA;
      YESNO (f_play_called_suit);
      COMMA;
      INT ((int) f_play_n_times_lead);
      COMMA;
      INT ((int) f_play_suit_remaining);
      COMMA;
      INT ((int) f_play_suit_remaining_points);
      COMMA;
      YESNO (f_ally_may_be_minor_master);
      COMMA;
      YESNO (f_ennemy_may_be_minor_master);
      COMMA;
      INT ((int) f_n_trumps_remaining);
      COMMA;
      YESNO (f_ally_may_be_trump_master);
      COMMA;
      YESNO (f_ennemy_may_be_trump_master);
      COMMA;
      INT ((int) f_n_halfpoints_in_trick);
      COMMA;
      INT ((int) f_n_oudlers_in_trick);
      switch (s_bid)
        {
        case TAROT_PASS:
          COMMA;
          STRING ("TAROT_PASS");
          break;
        case TAROT_TAKE:
          COMMA;
          STRING ("TAROT_TAKE");
          break;
        case TAROT_PUSH:
          COMMA;
          STRING ("TAROT_PUSH");
          break;
        case TAROT_STRAIGHT_KEEP:
          COMMA;
          STRING ("TAROT_STRAIGHT_KEEP");
          break;
        case TAROT_DOUBLE_KEEP:
          COMMA;
          STRING ("TAROT_DOUBLE_KEEP");
          break;
        default:
          assert (0);
        }
      COMMA;
      INT (s_outbid);
      COMMA;
      YESNO (s_call_shortest);
      COMMA;
      YESNO (s_call_second_shortest);
      COMMA;
      YESNO (s_call_second_longest);
      COMMA;
      YESNO (s_call_longest);
      COMMA;
      YESNO (s_call_myself);
      COMMA;
      YESNO (s_call_marriage);
      COMMA;
      INT ((int) s_discard_queens);
      COMMA;
      INT ((int) s_discard_knights);
      COMMA;
      INT ((int) s_discard_jacks);
      COMMA;
      INT ((int) s_discard_junk);
      COMMA;
      INT ((int) s_discard_shortest);
      COMMA;
      INT ((int) s_discard_second_shortest);
      COMMA;
      INT ((int) s_discard_second_longest);
      COMMA;
      INT ((int) s_discard_longest);
      COMMA;
      YESNO (s_lead_called_suit);
      COMMA;
      INT ((int) s_lead_suit_length);
      COMMA;
      INT ((int) s_lead_n_times_lead);
      COMMA;
      YESNO (s_lead_trumps);
      COMMA;
      INT ((int) s_n_smaller_kept);
      COMMA;
      INT ((int) s_n_greater_nonface_kept);
      COMMA;
      INT ((int) s_n_points_kept);
      COMMA;
      INT ((int) s_n_greater_kept);
      COMMA;
      YESNO (s_play_excuse);
      COMMA;
      YESNO (s_trump_master);
      COMMA;
      YESNO (s_trump_obvious_master);
      COMMA;
      INT ((int) s_halfpoints);
      COMMA;
      YESNO (s_oudler);
    }
  COMMA;
  INT (score);
  STRING ("\n");
  features_data_push ('\0', &max, &dest);
  return ret;
}

static void
features_variant (const TarotGame * game, size_t *n_players, int *with_call)
{
  *n_players = game_n_players (game);
  *with_call = game_with_call (game);
}

static void
features_position (const TarotGame * game, size_t *i_trick, size_t *position)
{
  TarotPlayer myself, leader;
  *i_trick = 0;
  if (game_get_main_player (game, &myself) != TAROT_GAME_OK)
    {
      *position = 0;
    }
  else
    {
      *position = (size_t) myself;
    }
  if (game_get_current_trick (game, i_trick) != TAROT_GAME_OK)
    {
      *i_trick = 0;
    }
  if (game_get_trick_leader (game, *i_trick, &leader) == TAROT_GAME_OK)
    {
      if (leader > *position)
        {
          *position += game_n_players (game);
        }
      assert (leader <= *position);
      *position -= leader;
    }
}

static void
features_partner_known (const TarotGame * game, int *known)
{
  TarotPlayer partner;
  *known = 0;
  if (!game_with_call (game)
      || game_get_partner (game, &partner) == TAROT_GAME_OK)
    {
      *known = 1;
    }
}

static void
features_bid_minimum (const TarotGame * game, TarotBid * minimum)
{
  TarotPlayer taker;
  size_t n_bids;
  game_get_taker (game, &taker);
  if (taker < game_n_players (game)
      && game_get_bids (game, &n_bids, taker, 1, minimum) == TAROT_GAME_OK)
    {
      if (n_bids == 0)
        {
          *minimum = TAROT_PASS;
        }
      else
        {
          /* minimum is set */
        }
    }
  else
    {
      *minimum = TAROT_PASS;
    }
}

static size_t
features_get_my_cards (const TarotGame * game, size_t max, TarotCard * cards)
{
  TarotPlayer myself;
  if (game_get_next (game, &myself) == TAROT_GAME_OK)
    {
      size_t n_cards;
      if (game_get_cards (game, myself, &n_cards, 0, max, cards) ==
          TAROT_GAME_OK)
        {
          size_t n_dog;
          TarotCard dog[6];
          TarotPlayer taker;
          game_get_taker (game, &taker);
          if (game_step (game) == TAROT_DOG
              && taker == myself
              && game_get_dog (game, &n_dog, 0, 6, dog) == TAROT_GAME_OK)
            {
              size_t i = 0;
              assert (n_dog <= 6);
              for (i = 0; i < n_dog; i++)
                {
                  if (n_cards < max)
                    {
                      cards[n_cards] = dog[i];
                    }
                  n_cards++;
                }
            }
          return n_cards;
        }
    }
  return 0;
}

static void
features_count_faces (const TarotGame * game,
                      size_t *n_kings,
                      size_t *n_marriages,
                      size_t *n_extended_marriages,
                      size_t *n_queens, size_t *n_knights)
{
  TarotCard my_cards[78];
  size_t n_cards = features_get_my_cards (game, 78, my_cards);
  size_t i;
  int has_knight[4] = { 0, 0, 0, 0 };
  int has_queen[4] = { 0, 0, 0, 0 };
  int has_king[4] = { 0, 0, 0, 0 };
  *n_kings = 0;
  *n_marriages = 0;
  *n_extended_marriages = 0;
  *n_queens = 0;
  *n_knights = 0;
  assert (n_cards <= 78);
  for (i = 0; i < n_cards; i++)
    {
      TarotNumber number;
      TarotSuit suit;
      if (decompose (my_cards[i], &number, &suit) == 0
          && suit != TAROT_TRUMPS && number >= TAROT_KNIGHT)
        {
          size_t suit_index = (size_t) suit;
          assert (suit_index < 4);
          switch (number)
            {
            case TAROT_KNIGHT:
              assert (!has_knight[suit_index]);
              has_knight[suit_index] = 1;
              break;
            case TAROT_QUEEN:
              assert (!has_queen[suit_index]);
              has_queen[suit_index] = 1;
              break;
            case TAROT_KING:
              assert (!has_king[suit_index]);
              has_king[suit_index] = 1;
              break;
            default:
              assert (0);
            }
        }
    }
  for (i = 0; i < 4; i++)
    {
      *n_kings += has_king[i];
      *n_marriages += has_king[i] && has_queen[i];
      *n_extended_marriages += has_king[i] && has_queen[i] && has_knight[i];
      *n_queens += has_queen[i];
      *n_knights += has_knight[i];
    }
}

static void
features_count_minor (const TarotGame * game,
                      size_t *n_shortest,
                      size_t *n_second_shortest,
                      size_t *n_second_longest, size_t *n_longest,
                      TarotSuit * suit_order)
{
  TarotCard my_cards[78];
  size_t n_cards = features_get_my_cards (game, 78, my_cards);
  size_t i;
  size_t n_by_suit[4] = { 0, 0, 0, 0 };
  size_t size_sorted;
  *n_shortest = 0;
  *n_second_shortest = 0;
  *n_second_longest = 0;
  *n_longest = 0;
  suit_order[0] = TAROT_HEARTS;
  suit_order[1] = TAROT_CLUBS;
  suit_order[2] = TAROT_DIAMONDS;
  suit_order[3] = TAROT_SPADES;
  assert (n_cards <= 78);
  for (i = 0; i < n_cards; i++)
    {
      TarotNumber number;
      TarotSuit suit;
      if (decompose (my_cards[i], &number, &suit) == 0
          && suit != TAROT_TRUMPS)
        {
          size_t suit_index = (size_t) suit;
          assert (suit_index < 4);
          n_by_suit[suit_index]++;
        }
    }
  for (size_sorted = 1; size_sorted < 4; size_sorted++)
    {
      size_t inserted = size_sorted;
      while (inserted > 0 && n_by_suit[inserted - 1] > n_by_suit[inserted])
        {
          size_t n_inserted = n_by_suit[inserted];
          TarotSuit suit_inserted = suit_order[inserted];
          n_by_suit[inserted] = n_by_suit[inserted - 1];
          n_by_suit[inserted - 1] = n_inserted;
          suit_order[inserted] = suit_order[inserted - 1];
          suit_order[inserted - 1] = suit_inserted;
          inserted--;
        }
    }
  for (i = 0; i + 1 < 4; i++)
    {
      assert (n_by_suit[i] <= n_by_suit[i + 1]);
    }
  *n_shortest = n_by_suit[0];
  *n_second_shortest = n_by_suit[1];
  *n_second_longest = n_by_suit[2];
  *n_longest = n_by_suit[3];
}

static void
features_count_oudlers (const TarotGame * game,
                        int *has_petit, int *has_excuse, int *has_twentyone)
{
  TarotCard my_cards[78];
  size_t n_cards = features_get_my_cards (game, 78, my_cards);
  size_t i;
  *has_petit = 0;
  *has_excuse = 0;
  *has_twentyone = 0;
  for (i = 0; i < n_cards; i++)
    {
      if (my_cards[i] == TAROT_PETIT)
        {
          *has_petit = 1;
        }
      else if (my_cards[i] == TAROT_EXCUSE)
        {
          *has_excuse = 1;
        }
      else if (my_cards[i] == TAROT_TWENTYONE)
        {
          *has_twentyone = 1;
        }
    }
}

static void
features_count_trumps (const TarotGame * game,
                       size_t *n_trumps,
                       TarotNumber * max_trump,
                       TarotNumber * min_trump,
                       TarotNumber * q1_trumps,
                       TarotNumber * q3_trumps, TarotNumber * median_trump)
{
  TarotCard my_cards[78];
  size_t n_cards = features_get_my_cards (game, 78, my_cards);
  size_t i;
  TarotNumber my_trumps[21];
  size_t my_n_trumps = 0;
  *max_trump = 0;
  *min_trump = 0;
  *q1_trumps = 0;
  *q3_trumps = 0;
  *median_trump = 0;
  for (i = 0; i < n_cards; i++)
    {
      TarotNumber number;
      TarotSuit suit;
      if (decompose (my_cards[i], &number, &suit) == 0
          && suit == TAROT_TRUMPS)
        {
          assert (my_n_trumps < 21);
          if (my_n_trumps > 0)
            {
              assert (my_trumps[my_n_trumps - 1] < number);
            }
          my_trumps[my_n_trumps++] = number;
        }
    }
  *n_trumps = my_n_trumps;
  if (my_n_trumps != 0)
    {
      double size_d = (double) my_n_trumps;
      double size_q1 = size_d / 4;
      double size_med = size_d / 2;
      double size_q3 = 3 * size_d / 4;
      *max_trump = my_trumps[my_n_trumps - 1];
      *min_trump = my_trumps[0];
      *q1_trumps = my_trumps[(size_t) size_q1];
      *q3_trumps = my_trumps[(size_t) size_q3];
      *median_trump = my_trumps[(size_t) size_med];
    }
}

static void
strategy_bid (const TarotGame * game,
              const TarotGameEvent * event, TarotBid * bid)
{
  (void) game;
  if (event_get_bid (event, bid) != TAROT_EVENT_OK)
    {
      *bid = TAROT_PASS;
    }
}

static void
strategy_outbid (const TarotGame * game,
                 const TarotGameEvent * event, int *outbid)
{
  TarotBid minimum;
  TarotBid actual;
  TarotPlayer next;
  *outbid = 0;
  if (game_get_hint_bid (game, &next, &minimum) == TAROT_GAME_OK
      && event_get_bid (event, &actual) == TAROT_EVENT_OK)
    {
      *outbid = ((int) actual) - ((int) minimum);
    }
}

static void
strategy_call_by_length (const TarotGame * game,
                         const TarotGameEvent * event,
                         int *shortest,
                         int *second_shortest,
                         int *second_longest, int *longest)
{
  TarotCard my_cards[78];
  size_t n_cards = features_get_my_cards (game, 78, my_cards);
  size_t i;
  size_t n_by_suit[4] = { 0, 0, 0, 0 };
  size_t n_fewer = 0;
  size_t n_more = 0;
  TarotSuit call_suit;
  TarotNumber call_number;
  TarotCard call;
  *shortest = 0;
  *second_shortest = 0;
  *second_longest = 0;
  *longest = 0;
  assert (n_cards <= 78);
  if (event_get_call (event, &call) != TAROT_EVENT_OK)
    {
      return;
    }
  if (decompose (call, &call_number, &call_suit) != 0)
    {
      return;
    }
  if (call_suit == TAROT_TRUMPS)
    {
      return;
    }
  for (i = 0; i < n_cards; i++)
    {
      TarotNumber number;
      TarotSuit suit;
      if (decompose (my_cards[i], &number, &suit) == 0
          && suit != TAROT_TRUMPS)
        {
          size_t suit_index = (size_t) suit;
          assert (suit_index < 4);
          n_by_suit[suit_index]++;
        }
    }
  n_fewer = n_more = 0;
  for (i = 0; i < 4; i++)
    {
      if (n_by_suit[i] < n_by_suit[(size_t) call_suit])
        {
          n_fewer++;
        }
      if (n_by_suit[i] > n_by_suit[(size_t) call_suit])
        {
          n_more++;
        }
    }
  switch (n_fewer)
    {
    case 0:
      *shortest = 1;
      break;
    case 1:
      *second_shortest = 1;
      break;
    case 2:
      *second_longest = 1;
      break;
    case 3:
      *longest = 1;
      break;
    default:
      assert (0);
    }
  switch (n_more)
    {
    case 0:
      *longest = 1;
      break;
    case 1:
      *second_longest = 1;
      break;
    case 2:
      *second_shortest = 1;
      break;
    case 3:
      *shortest = 1;
      break;
    default:
      assert (0);
    }
}

static void
strategy_call_myself (const TarotGame * game,
                      const TarotGameEvent * event, int *myself)
{
  TarotCard my_cards[78];
  size_t n_cards = features_get_my_cards (game, 78, my_cards);
  size_t i;
  TarotCard call;
  *myself = 0;
  assert (n_cards <= 78);
  if (event_get_call (event, &call) != TAROT_EVENT_OK)
    {
      return;
    }
  for (i = 0; i < n_cards; i++)
    {
      if (my_cards[i] == call)
        {
          *myself = 1;
        }
    }
}

static void
strategy_call_marriage (const TarotGame * game,
                        const TarotGameEvent * event, int *will_own_marriage)
{
  TarotCard my_cards[78];
  size_t n_cards = features_get_my_cards (game, 78, my_cards);
  size_t i;
  TarotCard call;
  TarotNumber call_number;
  TarotSuit call_suit;
  TarotNumber spouse_number;
  TarotCard spouse;
  *will_own_marriage = 0;
  assert (n_cards <= 78);
  if (event_get_call (event, &call) != TAROT_EVENT_OK)
    {
      return;
    }
  if (decompose (call, &call_number, &call_suit) != 0)
    {
      return;
    }
  switch (call_number)
    {
    case TAROT_KING:
      spouse_number = TAROT_QUEEN;
      break;
    case TAROT_QUEEN:
      spouse_number = TAROT_KING;
      break;
    default:
      return;
    }
  if (of (spouse_number, call_suit, &spouse) != 0)
    {
      assert (0);
    }
  assert (spouse == call + 1 || spouse + 1 == call);
  for (i = 0; i < n_cards; i++)
    {
      if (my_cards[i] == spouse)
        {
          *will_own_marriage = 1;
        }
    }
}

/* For each of the 78 cards, set its status to 0 by default, or +1 if
   we own that card, or +2 if we do not own it, do not discard it and
   call it, or -1 if we own it and discard it. */
static void
features_get_status (const TarotGame * game,
                     const TarotGameEvent * event, int *status)
{
  size_t i = 0;
  TarotCard my_cards[78];
  TarotCard discard[6];
  size_t n_cards = features_get_my_cards (game, 78, my_cards);
  size_t n_discard;
  TarotCard call;
  for (i = 0; i < 78; i++)
    {
      status[i] = 0;
    }
  assert (n_cards <= 78);
  if (event_get_discard (event, &n_discard, 0, 6, discard) != TAROT_EVENT_OK)
    {
      return;
    }
  assert (n_discard <= 6);
  for (i = 0; i < n_cards; i++)
    {
      status[my_cards[i]] = +1;
    }
  if (game_get_call (game, &call) == TAROT_GAME_OK)
    {
      status[call] = +2;
    }
  for (i = 0; i < n_discard; i++)
    {
      status[discard[i]] = -1;
    }
}

static void
strategy_discard_save (const TarotGame * game,
                       const TarotGameEvent * event,
                       size_t *queens, size_t *knights, size_t *jacks)
{
  int status[78] = { 0 };
  static const TarotSuit minor_suits[4] =
    { TAROT_HEARTS, TAROT_CLUBS, TAROT_DIAMONDS, TAROT_SPADES };
  size_t i;
  *queens = 0;
  *knights = 0;
  *jacks = 0;
  features_get_status (game, event, status);
  for (i = 0; i < 4; i++)
    {
      TarotSuit suit = minor_suits[i];
      TarotCard king, queen, knight, jack;
      if (of (TAROT_KING, suit, &king) != 0
          || of (TAROT_QUEEN, suit, &queen) != 0
          || of (TAROT_KNIGHT, suit, &knight) != 0
          || of (TAROT_JACK, suit, &jack) != 0)
        {
          assert (0);
        }
      if (status[king] == 0 && status[queen] < 0)
        {
          *queens += 1;
        }
      if ((status[king] == 0 || status[queen] == 0) && status[knight] < 0)
        {
          *knights += 1;
        }
      if ((status[king] == 0 || status[queen] == 0 || status[knight] == 0)
          && status[jack] < 0)
        {
          *jacks += 1;
        }
    }
}

static void
strategy_discard_junk (const TarotGame * game,
                       const TarotGameEvent * event, size_t *lesser)
{
  int status[78] = { 0 };
  static const TarotSuit minor_suits[4] =
    { TAROT_HEARTS, TAROT_CLUBS, TAROT_DIAMONDS, TAROT_SPADES };
  size_t i;
  *lesser = 0;
  features_get_status (game, event, status);
  for (i = 0; i < 4; i++)
    {
      TarotSuit suit = minor_suits[i];
      TarotCard ace, ten, c;
      if (of (TAROT_ACE, suit, &ace) != 0 || of (10, suit, &ten) != 0)
        {
          assert (0);
        }
      /* Check if we discarded any minor card of that suit */
      for (c = ace; c <= ten && status[c] >= 0; c++);
      if (c <= ten)
        {
          for (c = ace; c <= ten && status[c] >= 0; c++)
            {
              if (status[c] > 0)
                {
                  *lesser += 1;
                }
            }
        }
    }
}

static void
strategy_discard_by_length (const TarotGame * game,
                            const TarotGameEvent * event,
                            size_t *shortest,
                            size_t *second_shortest,
                            size_t *second_longest, size_t *longest)
{
  TarotSuit suit_order[4];
  size_t i = 0;
  TarotCard discard[6];
  size_t n_discard;
  size_t n_shortest;
  size_t n_second_shortest;
  size_t n_second_longest;
  size_t n_longest;
  *shortest = 0;
  *second_shortest = 0;
  *second_longest = 0;
  *longest = 0;
  features_count_minor (game, &n_shortest, &n_second_shortest,
                        &n_second_longest, &n_longest, suit_order);
  if (event_get_discard (event, &n_discard, 0, 6, discard) != TAROT_EVENT_OK)
    {
      return;
    }
  assert (n_discard <= 6);
  for (i = 0; i < n_discard; i++)
    {
      TarotNumber n;
      TarotSuit s;
      if (decompose (discard[i], &n, &s) != 0)
        {
          assert (0);
        }
      if (s == suit_order[0])
        {
          *shortest += 1;
        }
      else if (s == suit_order[1])
        {
          *second_shortest += 1;
        }
      else if (s == suit_order[2])
        {
          *second_longest += 1;
        }
      else
        {
          assert (s == suit_order[3]);
          *longest += 1;
        }
    }
}

static void
get_teams (const TarotGame * game, int *attack)
{
  size_t n_players = game_n_players (game), i;
  TarotPlayer taker, partner;
  assert (n_players <= 5);
  game_get_taker (game, &taker);
  if (game_get_partner (game, &partner) != TAROT_GAME_OK)
    {
      partner = taker;
    }
  for (i = 0; i < n_players; i++)
    {
      attack[i] = 0;
    }
  attack[taker] = 1;
  attack[partner] = 1;
}

static void
features_team (const TarotGame * game, int *attack, int *defence)
{
  TarotPlayer myself;
  int teams[5];
  *attack = 0;
  *defence = 0;
  get_teams (game, teams);
  if (tarot_game_get_main_player (game, &myself) == TAROT_GAME_OK)
    {
      *attack = teams[myself];
      *defence = 1 - teams[myself];
    }
}

static void
features_trick_players (const TarotGame * game,
                        size_t *n_already_played,
                        size_t *n_allies_after_last_opponent)
{
  size_t i_trick;
  size_t n_players = game_n_players (game), i, n_cards;
  TarotPlayer leader, myself;
  int teams[5];
  TarotPlayer partner;
  *n_already_played = 0;
  *n_allies_after_last_opponent = 0;
  if (game_get_current_trick (game, &i_trick) != TAROT_GAME_OK
      || game_get_main_player (game, &myself) != TAROT_GAME_OK)
    {
      return;
    }
  if (game_get_trick_leader (game, i_trick, &leader) != TAROT_GAME_OK)
    {
      return;
    }
  assert (leader < n_players);
  if (game_get_trick_cards (game, i_trick, &n_cards, 0, 0, NULL) ==
      TAROT_GAME_OK)
    {
      *n_already_played = n_cards;
    }
  get_teams (game, teams);
  assert (n_players >= 3 && n_players <= 5);
  if (!game_with_call (game)
      || game_get_partner (game, &partner) == TAROT_GAME_OK)
    {
      size_t last_opponent;     /* Number of players that have played before
                                 * the last opponent */
      last_opponent = n_cards;
      for (i = n_cards + 1; i < n_players; i++)
        {
          TarotPlayer player = (leader + i) % n_players;
          int team = teams[player];
          int is_ally = (team == teams[myself]);
          if (!is_ally)
            {
              last_opponent = i;
            }
        }
      *n_allies_after_last_opponent = n_players - last_opponent - 1;
    }
}

static void
may_play_in_trick (const TarotCounter * counter,
                   size_t n_players,
                   TarotPlayer first_forced,
                   size_t n_forced,
                   const TarotCard * forced_cards,
                   TarotCard candidate, int *may_play)
{
  size_t i;
  assert (n_players <= 5);
  for (i = 0; i < n_players; i++)
    {
      TarotPlayer player = (i + first_forced) % n_players;
      if (i < n_forced)
        {
          may_play[player] = (candidate == forced_cards[i]);
        }
      else
        {
          may_play[player] = counter_may_own (counter, 1, &player, candidate);
        }
    }
}

static int
find_master (const TarotGame * game,
             const TarotCounter * counter,
             TarotCard least_candidate,
             TarotCard strongest_candidate, int *may_play)
{
  size_t n_players = game_n_players (game);
  size_t i_trick;
  size_t i_candidate;
  TarotGameCardAnalysis cards_full[5];
  TarotCard cards[5];
  size_t n_cards, i_card;
  TarotPlayer leader;
  if (game_get_current_trick (game, &i_trick) != TAROT_GAME_OK)
    {
      return 1;
    }
  if (game_get_trick_leader (game, i_trick, &leader) != TAROT_GAME_OK)
    {
      return 1;
    }
  if (game_get_trick_cards (game, i_trick, &n_cards, 0, 5, cards_full) !=
      TAROT_GAME_OK)
    {
      return 1;
    }
  for (i_card = 0; i_card < n_cards; i_card++)
    {
      cards[i_card] = cards_full[i_card].card;
    }
  for (i_candidate = 0;
       i_candidate < strongest_candidate - least_candidate + 1; i_candidate++)
    {
      TarotCard candidate = strongest_candidate - i_candidate;
      TarotPlayer i;
      may_play_in_trick (counter, n_players, leader, n_cards, cards,
                         candidate, may_play);
      for (i = 0; i < n_players; i++)
        {
          if (may_play[i])
            {
              return 0;
            }
        }
    }
  return 0;
}

static int
find_master_team (const TarotGame * game,
                  const TarotCounter * counter,
                  TarotCard least_candidate,
                  TarotCard strongest_candidate, int *ally, int *ennemy)
{
  int teams[5];
  int may_play[5] = { 0 };
  size_t i;
  TarotPlayer myself;
  get_teams (game, teams);
  *ally = 0;
  *ennemy = 0;
  if (find_master
      (game, counter, least_candidate, strongest_candidate, may_play) != 0)
    {
      return 1;
    }
  if (game_get_main_player (game, &myself) != 0)
    {
      return 1;
    }
  /* Since may_play has been initialized with extra zeros, this works
   * with 3 or 4 players too */
  for (i = 0; i < 5; i++)
    {
      if (may_play[i])
        {
          if (teams[i] == teams[myself])
            {
              *ally = 1;
            }
          else
            {
              *ennemy = 1;
            }
        }
    }
  return 0;
}

static int
get_called_suit (const TarotGame * game, TarotSuit * called)
{
  TarotCard call;
  if (game_get_call (game, &call) == TAROT_GAME_OK)
    {
      TarotNumber n;
      if (decompose (call, &n, called) != 0)
        {
          assert (0);
        }
      return 0;
    }
  return 1;
}

static size_t
count_times_lead (const TarotGame * game, TarotSuit suit)
{
  size_t current_trick;
  size_t trick;
  size_t ret = 0;
  if (game_get_current_trick (game, &current_trick) != TAROT_GAME_OK)
    {
      /* Fun fact: this is actually correct! */
      return 0;
    }
  for (trick = 0; trick <= current_trick; trick++)
    {
      TarotSuit lead_suit;
      if (game_get_lead_suit (game, trick, &lead_suit) == TAROT_GAME_OK)
        {
          ret += (suit == lead_suit);
        }
    }
  return ret;
}

static void
features_minor_suit (const TarotGame * game,
                     const TarotCounter * counter,
                     int *called_suit,
                     size_t *n_times_lead,
                     size_t *n_remaining_cards,
                     size_t *n_remaining_points,
                     int *ally_may_be_master, int *ennemy_may_be_master)
{
  TarotSuit the_called_suit, lead_suit;
  size_t i_trick;
  TarotCard first_of_lead_suit;
  TarotCard last_of_lead_suit;
  TarotCard card_of_lead_suit;
  size_t n_players = tarot_game_n_players (game);
  *called_suit = 0;
  *n_times_lead = 0;
  *n_remaining_cards = 0;
  *n_remaining_points = 0;
  *ally_may_be_master = 0;
  *ennemy_may_be_master = 0;
  assert (n_players <= 5);
  if (game_get_current_trick (game, &i_trick) != TAROT_GAME_OK)
    {
      return;
    }
  if (game_get_lead_suit (game, i_trick, &lead_suit) != TAROT_GAME_OK)
    {
      return;
    }
  if (get_called_suit (game, &the_called_suit) == 0)
    {
      *called_suit = (lead_suit == the_called_suit);
    }
  *n_times_lead = count_times_lead (game, lead_suit);
  if (of (1, lead_suit, &first_of_lead_suit) != 0)
    {
      assert (0);
    }
  if (lead_suit == TAROT_TRUMPS)
    {
      if (of (21, lead_suit, &last_of_lead_suit) != 0)
        {
          assert (0);
        }
    }
  else
    {
      if (of (TAROT_KING, lead_suit, &last_of_lead_suit) != 0)
        {
          assert (0);
        }
    }
  for (card_of_lead_suit = first_of_lead_suit;
       card_of_lead_suit <= last_of_lead_suit; card_of_lead_suit++)
    {
      TarotPlayer current_owner =
        counter->current_owner_of[card_of_lead_suit];
      static const size_t points[78] = {
        1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 3, 5, 7, 9,
        1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 3, 5, 7, 9,
        1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 3, 5, 7, 9,
        1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 3, 5, 7, 9,
        9, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 9,
        1
      };
      if (current_owner == TAROT_COUNTER_CARD_OWNER_PLAYED
          || current_owner == TAROT_COUNTER_CARD_OWNER_DISCARDED)
        {
          /* This card is not available */
        }
      else
        {
          *n_remaining_cards += 1;
          *n_remaining_points += points[card_of_lead_suit];
        }
    }
  assert (*n_remaining_points <= 37);
  if (find_master_team
      (game, counter, first_of_lead_suit, last_of_lead_suit,
       ally_may_be_master, ennemy_may_be_master) != 0)
    {
      *ally_may_be_master = 0;
      *ennemy_may_be_master = 0;
    }
}

static void
features_trump (const TarotGame * game,
                const TarotCounter * counter,
                size_t *n_remaining,
                int *ally_may_be_master, int *ennemy_may_be_master)
{
  TarotCard first = TAROT_PETIT;
  TarotCard last = TAROT_TWENTYONE;
  TarotCard card;
  size_t n_players = tarot_game_n_players (game);
  *n_remaining = 0;
  *ally_may_be_master = 0;
  *ennemy_may_be_master = 0;
  assert (n_players <= 5);
  for (card = first; card <= last; card++)
    {
      TarotPlayer current_owner = counter->current_owner_of[card];
      if (current_owner == TAROT_COUNTER_CARD_OWNER_PLAYED
          || current_owner == TAROT_COUNTER_CARD_OWNER_DISCARDED)
        {
          /* This card is not available */
        }
      else
        {
          *n_remaining += 1;
        }
    }
  if (find_master_team
      (game, counter, first, last, ally_may_be_master,
       ennemy_may_be_master) != 0)
    {
      *ally_may_be_master = 0;
      *ennemy_may_be_master = 0;
    }
}

static void
features_points (const TarotGame * game,
                 size_t *n_halfpoints, size_t *n_oudlers)
{
  size_t n_cards;
  TarotGameCardAnalysis cards[5];
  size_t i_trick;
  size_t i_card;
  *n_oudlers = 0;
  *n_halfpoints = 0;
  if (game_get_current_trick (game, &i_trick) != TAROT_GAME_OK)
    {
      return;
    }
  if (game_get_trick_cards (game, i_trick, &n_cards, 0, 5, cards) !=
      TAROT_GAME_OK)
    {
      return;
    }
  assert (n_cards <= 5);
  for (i_card = 0; i_card < n_cards; i_card++)
    {
      static const unsigned int halfpoints[78] = {
        1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 3, 5, 7, 9,
        1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 3, 5, 7, 9,
        1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 3, 5, 7, 9,
        1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 3, 5, 7, 9,
        9, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 9,
        1
      };
      static const int is_oudler[78] = {
        0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0,
        0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0,
        0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0,
        0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0,
        1, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 1,
        0
      };
      *n_oudlers += is_oudler[cards[i_card].card];
      *n_halfpoints += halfpoints[cards[i_card].card];
    }
}

static void
strategy_lead (const TarotGame * game,
               const TarotCounter * counter,
               const TarotGameEvent * event,
               int *called_suit,
               size_t *suit_length, size_t *n_times_lead, int *trumps)
{
  TarotCard card;
  int is_decomposable;
  TarotSuit card_suit;
  TarotNumber card_number;
  (void) counter;
  *called_suit = 0;
  *suit_length = 0;
  *n_times_lead = 0;
  *trumps = 0;
  if (event_get_card (event, &card) != TAROT_EVENT_OK)
    {
      return;
    }
  is_decomposable = (decompose (card, &card_number, &card_suit) == 0);
  if (is_decomposable)
    {
      TarotCard my_cards[78];
      size_t n_my_cards =
        features_get_my_cards (game, sizeof (my_cards) / sizeof (*my_cards),
                               my_cards);
      size_t i_card;
      TarotCard first;
      TarotCard last;
      TarotNumber first_number = 1;
      TarotNumber last_number = TAROT_KING;
      size_t n_of_suit = 0;
      TarotSuit call_suit;
      if (card_suit == TAROT_TRUMPS)
        {
          last_number = 21;
        }
      if (of (first_number, card_suit, &first) != 0
          || of (last_number, card_suit, &last) != 0)
        {
          assert (0);
        }
      for (i_card = 0; i_card < n_my_cards; i_card++)
        {
          if (my_cards[i_card] >= first && my_cards[i_card] <= last)
            {
              n_of_suit++;
            }
        }
      if (get_called_suit (game, &call_suit) == 0)
        {
          *called_suit = (call_suit == card_suit);
        }
      *suit_length = n_of_suit;
      *n_times_lead = count_times_lead (game, card_suit);
      *trumps = (card_suit == TAROT_TRUMPS);
    }
}

static void
strategy_purity (const TarotGame * game,
                 const TarotGameEvent * event,
                 size_t *n_smaller_kept,
                 size_t *n_greater_nonface_kept,
                 size_t *n_points_kept, size_t *n_greater_kept)
{
  TarotCard card;
  int is_decomposable;
  TarotSuit card_suit;
  TarotNumber card_number;
  *n_smaller_kept = 0;
  *n_greater_nonface_kept = 0;
  *n_points_kept = 0;
  *n_greater_kept = 0;
  if (event_get_card (event, &card) != TAROT_EVENT_OK)
    {
      return;
    }
  is_decomposable = (decompose (card, &card_number, &card_suit) == 0);
  if (is_decomposable)
    {
      TarotCard my_cards[78];
      size_t n_my_cards =
        features_get_my_cards (game, sizeof (my_cards) / sizeof (*my_cards),
                               my_cards);
      size_t i_card;
      TarotCard first;
      TarotCard last_nonface;
      TarotCard last;
      TarotNumber first_number = 1;
      TarotNumber last_nonface_number = 10;
      TarotNumber last_number = TAROT_KING;
      size_t n_lesser = 0;
      size_t n_greater_nonface = 0;
      size_t n_greater = 0;
      size_t n_points = 0;
      if (card_suit == TAROT_TRUMPS)
        {
          last_nonface_number = 21;
          last_number = 21;
        }
      if (of (first_number, card_suit, &first) != 0
          || of (last_nonface_number, card_suit, &last_nonface) != 0
          || of (last_number, card_suit, &last) != 0)
        {
          assert (0);
        }
      for (i_card = 0; i_card < n_my_cards; i_card++)
        {
          if (my_cards[i_card] >= first && my_cards[i_card] < card)
            {
              n_lesser++;
            }
          if (my_cards[i_card] > card && my_cards[i_card] <= last_nonface)
            {
              n_greater_nonface++;
            }
          if (my_cards[i_card] > card && my_cards[i_card] <= last)
            {
              n_greater++;
            }
          if (my_cards[i_card] >= first && my_cards[i_card] <= last
              && my_cards[i_card] != card)
            {
              static const unsigned int halfpoints[78] = {
                1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 3, 5, 7, 9,
                1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 3, 5, 7, 9,
                1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 3, 5, 7, 9,
                1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 3, 5, 7, 9,
                9, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 9,
                1
              };
              n_points += halfpoints[my_cards[i_card]];
            }
        }
      *n_smaller_kept = n_lesser;
      *n_greater_nonface_kept = n_greater_nonface;
      *n_greater_kept = n_greater;
      *n_points_kept = n_points;
    }
}

static void
strategy_excuse (const TarotGame * game,
                 const TarotGameEvent * event, int *excuse)
{
  TarotCard card;
  (void) game;
  *excuse = 0;
  if (event_get_card (event, &card) == TAROT_EVENT_OK)
    {
      *excuse = (card == TAROT_EXCUSE);
    }
}

static size_t
who_plays_after_me (const TarotGame * game, TarotPlayer * p)
{
  size_t i_trick, ret, i, n_cards, n_players = game_n_players (game);
  TarotPlayer leader;
  if (game_get_current_trick (game, &i_trick) != TAROT_GAME_OK)
    {
      return 0;
    }
  if (game_get_trick_leader (game, i_trick, &leader) != TAROT_GAME_OK)
    {
      return 0;
    }
  if (game_get_trick_cards (game, i_trick, &n_cards, 0, 0, NULL) !=
      TAROT_GAME_OK)
    {
      return 0;
    }
  ret = 0;
  for (i = n_cards + 1; i < n_players; i++)
    {
      assert (ret <= 5);
      p[ret++] = (leader + i) % n_players;
    }
  return ret;
}

static void
strategy_trump_master (const TarotGame * game,
                       const TarotCounter * counter,
                       const TarotGameEvent * event,
                       int *actually_master, int *obvious_master)
{
  TarotCard card, c;
  TarotNumber number, number_max_trump;
  TarotSuit suit;
  size_t n_players = game_n_players (game);
  TarotPlayer after_me[5];
  size_t n_after_me = who_plays_after_me (game, after_me);
  size_t current_trick;
  *actually_master = 0;
  *obvious_master = 0;
  if (event_get_card (event, &card) != TAROT_EVENT_OK)
    {
      return;
    }
  if (decompose (card, &number, &suit) != 0)
    {
      return;
    }
  if (suit != TAROT_TRUMPS)
    {
      return;
    }
  if (game_get_current_trick (game, &current_trick) != TAROT_GAME_OK)
    {
      return;
    }
  if (game_get_max_trump (game, current_trick, &number_max_trump) ==
      TAROT_GAME_OK)
    {
      if (number_max_trump > number)
        {
          /* Someone has already played better than us in this
           * trick. */
          return;
        }
    }
  *obvious_master = 1;
  *actually_master = 1;
  for (c = card + 1; c < TAROT_EXCUSE; c++)
    {
      if (counter->current_owner_of[c] < n_players)
        {
          *obvious_master = 0;
        }
      if (counter_may_own (counter, n_after_me, after_me, c))
        {
          *actually_master = 0;
          *obvious_master = 0;
        }
    }
}

static void
strategy_play_points (const TarotGame * game,
                      const TarotGameEvent * event,
                      size_t *n_halfpoints, int *oudler)
{
  TarotCard card;
  static const unsigned int halfpoints[78] = {
    1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 3, 5, 7, 9,
    1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 3, 5, 7, 9,
    1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 3, 5, 7, 9,
    1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 3, 5, 7, 9,
    9, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 9,
    1
  };
  static const int is_oudler[78] = {
    0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0,
    0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0,
    0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0,
    0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0,
    1, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 1,
    0
  };
  (void) game;
  *n_halfpoints = 0;
  *oudler = 0;
  if (event_get_card (event, &card) != TAROT_EVENT_OK)
    {
      return;
    }
  *n_halfpoints = halfpoints[card];
  *oudler = is_oudler[card];
}

static double
features_strongest_player_try_bid (const TarotGame * game)
{
  double scores[4];
  TarotGameEvent candidates[4];
  TarotGameEvent *pointers[4];
  size_t i;
  double max = -DBL_MAX;
  TarotPerceptron *perceptron = tarot_perceptron_static_default ();
  for (i = 0; i < 4; i++)
    {
      candidates[i].t = TAROT_BID_EVENT;
      pointers[i] = &(candidates[i]);
    }
  candidates[0].u.bid = TAROT_TAKE;
  candidates[1].u.bid = TAROT_PUSH;
  candidates[2].u.bid = TAROT_STRAIGHT_KEEP;
  candidates[3].u.bid = TAROT_DOUBLE_KEEP;
  tarot_perceptron_eval (perceptron, game, 4, pointers, 0, 4, scores);
  for (i = 0; i < 4; i++)
    {
      if (scores[i] > max)
        {
          max = scores[i];
        }
    }
  tarot_perceptron_free (perceptron);
  return max;
}

static double
features_strongest_player_aux (size_t n_players,
                               int with_call,
                               size_t n_cards,
                               const TarotPlayer * owners, TarotPlayer who)
{
  TarotGame game;
  TarotGameEvent deal_all;
  TarotPlayer my_owners[78];
  size_t i;
  assert (n_cards == 78);
  for (i = 0; i < 78; i++)
    {
      if (owners[i] >= n_players)
        {
          my_owners[i] = n_players;
        }
      else
        {
          my_owners[i] = (owners[i] + n_players - who) % n_players;
        }
    }
  deal_all.t = TAROT_DEAL_ALL_EVENT;
  deal_all.n = 78;
  deal_all.data = my_owners;
  game_initialize (&game, n_players, with_call);
  if (game_add_event (&game, &deal_all) != TAROT_GAME_OK)
    {
      /* Petit sec */
      return -DBL_MAX;
    }
  return features_strongest_player_try_bid (&game);
}

static int predictor_perceptron_select_call (const TarotGame * game,
                                             TarotCard * call, double *score);

static inline TarotCard
features_strongest_call (size_t n_players, size_t n_cards,
                         const TarotPlayer * owners, TarotPlayer taker)
{
  /* Anything may fail in this function if there is a petit sec.  We
   * need to fail gracefully. */
  TarotGame game;
  TarotCard call;
  double score;
  TarotGameEvent setup;
  size_t i;
  game_initialize (&game, n_players, 1);
  setup.t = TAROT_SETUP_EVENT;
  setup.u.setup.n_players = n_players;
  setup.u.setup.with_call = 1;
  if (game_add_event (&game, &setup) != TAROT_GAME_OK)
    {
      return ((TarotCard) (-1));
    }
  if (game_duplicate_set_deal (&game, n_cards, owners) != TAROT_GAME_OK)
    {
      /* Anyway, the call event will never be evaluated. */
      return ((TarotCard) (-1));
    }
  if (game_duplicate_set_taker (&game, taker) != TAROT_GAME_OK)
    {
      /* Since we have dealt, we cannot fail here */
      assert (0);
    }
  for (i = 0; i < n_players; i++)
    {
      TarotGameEvent decl;
      decl.t = TAROT_DECL_EVENT;
      decl.u.decl = 0;
      /* This one cannot fail */
      if (game_add_event (&game, &decl) != TAROT_GAME_OK)
        {
          assert (0);
        }
    }
  if (predictor_perceptron_select_call (&game, &call, &score) != 0)
    {
      assert (0);
    }
  return call;
}

static int
predictor_perceptron_select_call (const TarotGame * game, TarotCard * call,
                                  double *best_score)
{
  static const TarotSuit suits[4] = {
    TAROT_HEARTS,
    TAROT_CLUBS,
    TAROT_DIAMONDS,
    TAROT_SPADES
  };
  TarotNumber minimum, candidate;
  TarotGameEvent e;
  TarotGameEvent *pointer_e = &e;
  TarotPlayer player;
  int ret = 1;
  int isuit;
  TarotPerceptron *perceptron;
  e.n = 0;
  *best_score = -DBL_MAX;
  if (game_get_hint_call (game, &player, &minimum) != TAROT_GAME_OK)
    {
      return 1;
    }
  perceptron = tarot_perceptron_static_default ();
  for (isuit = 0; isuit < 4; isuit++)
    {
      TarotSuit suit = suits[isuit];
      for (candidate = minimum; candidate <= TAROT_KING; candidate++)
        {
          TarotCard c;
          double candidate_score;
          if (tarot_of (candidate, suit, &c) != 0)
            {
              assert (0);
            }
          event_set_call (&e, c);
          tarot_perceptron_eval (perceptron, game, 1, &pointer_e, 0, 1,
                                 &candidate_score);
          if (candidate_score > *best_score)
            {
              *call = c;
              *best_score = candidate_score;
              ret = 0;
            }
        }
    }
  tarot_perceptron_free (perceptron);
  return ret;
}

static inline TarotPlayer
features_strongest_player (size_t n_players,
                           int with_call,
                           size_t n_cards,
                           const TarotPlayer * owners, double *confidence,
                           TarotCard * call)
{
  TarotPlayer candidate;
  TarotPlayer best = 0;
  double candidate_score;
  double best_score =
    features_strongest_player_aux (n_players, with_call, n_cards, owners, 0);
  for (candidate = 1; candidate < n_players; candidate++)
    {
      candidate_score =
        features_strongest_player_aux (n_players, with_call, n_cards, owners,
                                       candidate);
      if (candidate_score > best_score)
        {
          best = candidate;
          best_score = candidate_score;
        }
    }
  *confidence = best_score;
  if (with_call)
    {
      *call = features_strongest_call (n_players, n_cards, owners, best);
    }
  return best;
}

static inline double
features_errors_client (const TarotGame * game, int just_bids)
{
  double ret = 0;
  TarotGameIterator it;
  const TarotGameEvent *event;
  TarotPredictor predictor;
  size_t alignment;
  TarotGame rebuilt;
  TarotPlayer main_player;
  size_t n_bids_seen = 0;
  size_t n_players = game_n_players (game);
  game_iterator_setup (&it, game);
  game_initialize (&rebuilt, n_players, game_with_call (game));
  if (game_get_main_player (game, &main_player) != TAROT_GAME_OK)
    {
      assert (0);
    }
  if (predictor_construct
      (sizeof (predictor), (char *) &predictor,
       &alignment) != sizeof (predictor))
    {
      assert (0);
    }
  while ((event = game_iterator_next_value (&it)) != NULL)
    {
      TarotPlayer next;
      if (!just_bids || n_bids_seen < n_players)
        {
          if (game_get_next (&rebuilt, &next) == TAROT_GAME_OK)
            {
              if (next == main_player)
                {
                  double best_score, score;
                  predictor_set_game (&predictor, &rebuilt);
                  predictor_best (&predictor, &best_score);
                  if (predictor_eval
                      (&predictor, 1, (TarotGameEvent **) (&event), 0, 1,
                       &score) != 1)
                    {
                      assert (0);
                    }
                  if (score <= best_score)
                    {
                      /* It is possible that the predictor missed this
                       * one. */
                      ret += best_score - score;
                    }
                }
            }
          if (game_add_event (&rebuilt, event) != TAROT_GAME_OK)
            {
              assert (0);
            }
        }
    }
  return ret;
}

static inline double
features_errors_aux (const TarotGame * game, int just_bids)
{
  TarotPlayer main;
  size_t i, n = game_n_players (game);
  double ret = 0;
  if (game_get_main_player (game, &main) == TAROT_GAME_OK)
    {
      return features_errors_client (game, just_bids);
    }
  else if (game_step (game) <= TAROT_DEAL)
    {
      return 0;
    }
  for (i = 0; i < n; i++)
    {
      TarotGame g;
      game_initialize (&g, n, game_with_call (game));
      game_copy_as (&g, game, i);
      ret += features_errors_client (&g, just_bids);
    }
  return (ret / n);
}

static inline double
features_errors (const TarotGame * game)
{
  return features_errors_aux (game, 0);
}

static inline double
features_errors_just_bids (const TarotGame * game)
{
  return features_errors_aux (game, 1);
}

static size_t
predictor_construct (size_t max_mem, char *mem, size_t *alignment)
{
  size_t required = sizeof (TarotPredictor);
  *alignment = alignof (TarotPredictor);
  if (max_mem >= required)
    {
      TarotPredictor *predictor = (TarotPredictor *) mem;
      game_initialize (&(predictor->game), 4, 0);
      predictor->can_play = 0;
      predictor->best_event.n = 6;
      predictor->best_event.data = predictor->event_data;
      predictor->cnn = 0;
      return max_mem;
    }
  return required;
}

static inline void
predictor_copy (TarotPredictor * dest, const TarotPredictor * source)
{
  game_copy (&(dest->game), &(source->game));
  dest->can_play = source->can_play;
  if (source->can_play)
    {
      event_copy (&(dest->best_event), &(source->best_event));
    }
}

static double
stacking_predictor_eval_one (const TarotGame * game,
                             const TarotGameEvent * event)
{
  size_t n_players;
  int with_call;
  size_t trick;
  size_t position;
  TarotGameEventT type = event_type (event);
  int partner_known;
  TarotBid bid_minimum;
  size_t bid_n_kings;
  size_t bid_n_marriages;
  size_t bid_n_extended_marriages;
  size_t bid_n_queens;
  size_t bid_n_knights;
  size_t bid_n_shortest;
  size_t bid_n_second_shortest;
  size_t bid_n_second_longest;
  size_t bid_n_longest;
  int bid_has_petit;
  int bid_has_excuse;
  int bid_has_twentyone;
  size_t bid_n_trumps;
  TarotNumber bid_max_trump;
  TarotNumber bid_min_trump;
  TarotNumber bid_q1_trumps;
  TarotNumber bid_q3_trumps;
  TarotNumber bid_median_trump;
  int team_attack;
  int team_defence;
  size_t n_trick_players_already_played;
  size_t n_allies_after_last_opponent;
  int play_called_suit;
  size_t play_n_times_lead;
  size_t play_suit_remaining;
  size_t play_suit_remaining_points;
  int ally_may_be_minor_master;
  int ennemy_may_be_minor_master;
  size_t n_trumps_remaining;
  int ally_may_be_trump_master;
  int ennemy_may_be_trump_master;
  size_t n_halfpoints_in_trick;
  size_t n_oudlers_in_trick;
  TarotBid strategy_bid_v;
  int strategy_outbid_v;
  int strategy_call_shortest;
  int strategy_call_second_shortest;
  int strategy_call_second_longest;
  int strategy_call_longest;
  int strategy_call_myself_v;
  int strategy_call_marriage_v;
  size_t strategy_discard_save_queens;
  size_t strategy_discard_save_knights;
  size_t strategy_discard_save_jacks;
  size_t strategy_discard_save_junk;
  size_t strategy_discard_shortest;
  size_t strategy_discard_second_shortest;
  size_t strategy_discard_second_longest;
  size_t strategy_discard_longest;
  int strategy_lead_called_suit;
  size_t strategy_lead_suit_length;
  size_t strategy_lead_suit_n_times_lead;
  int strategy_lead_trumps;
  size_t strategy_n_smaller_kept;
  size_t strategy_n_greater_nonface_kept;
  size_t strategy_n_points_kept;
  size_t strategy_n_greater_kept;
  int strategy_play_excuse;
  int strategy_trump_master_v;
  int strategy_trump_obvious_master;
  size_t strategy_n_halfpoints;
  int strategy_oudler;
  double score = 0;
  double tree_1, tree_2, tree_3, tree_4, tree_5, linear_1, linear_2, linear_3,
    linear_4, linear_5;
  TarotSuit order[4];
  TarotCounter counter;
  counter_initialize (&counter);
  counter_load_from_game (&counter, game);
  features_variant (game, &n_players, &with_call);
  features_position (game, &trick, &position);
  features_partner_known (game, &partner_known);
  features_bid_minimum (game, &bid_minimum);
  features_count_faces (game, &bid_n_kings, &bid_n_marriages,
                        &bid_n_extended_marriages, &bid_n_queens,
                        &bid_n_knights);
  features_count_minor (game, &bid_n_shortest, &bid_n_second_shortest,
                        &bid_n_second_longest, &bid_n_longest, order);
  features_count_oudlers (game, &bid_has_petit, &bid_has_excuse,
                          &bid_has_twentyone);
  features_count_trumps (game, &bid_n_trumps, &bid_max_trump, &bid_min_trump,
                         &bid_q1_trumps, &bid_q3_trumps, &bid_median_trump);
  features_team (game, &team_attack, &team_defence);
  features_trick_players (game, &n_trick_players_already_played,
                          &n_allies_after_last_opponent);
  features_minor_suit (game, &counter, &play_called_suit, &play_n_times_lead,
                       &play_suit_remaining, &play_suit_remaining_points,
                       &ally_may_be_minor_master,
                       &ennemy_may_be_minor_master);
  features_trump (game, &counter, &n_trumps_remaining,
                  &ally_may_be_trump_master, &ennemy_may_be_trump_master);
  features_points (game, &n_halfpoints_in_trick, &n_oudlers_in_trick);
  strategy_bid (game, event, &strategy_bid_v);
  strategy_outbid (game, event, &strategy_outbid_v);
  strategy_call_by_length (game, event, &strategy_call_shortest,
                           &strategy_call_second_shortest,
                           &strategy_call_second_longest,
                           &strategy_call_longest);
  strategy_call_myself (game, event, &strategy_call_myself_v);
  strategy_call_marriage (game, event, &strategy_call_marriage_v);
  strategy_discard_save (game, event, &strategy_discard_save_queens,
                         &strategy_discard_save_knights,
                         &strategy_discard_save_jacks);
  strategy_discard_junk (game, event, &strategy_discard_save_junk);
  strategy_discard_by_length (game, event, &strategy_discard_shortest,
                              &strategy_discard_second_shortest,
                              &strategy_discard_second_longest,
                              &strategy_discard_longest);
  strategy_lead (game, &counter, event, &strategy_lead_called_suit,
                 &strategy_lead_suit_length, &strategy_lead_suit_n_times_lead,
                 &strategy_lead_trumps);
  strategy_purity (game, event, &strategy_n_smaller_kept,
                   &strategy_n_greater_nonface_kept, &strategy_n_points_kept,
                   &strategy_n_greater_kept);
  strategy_excuse (game, event, &strategy_play_excuse);
  strategy_trump_master (game, &counter, event, &strategy_trump_master_v,
                         &strategy_trump_obvious_master);
  strategy_play_points (game, event, &strategy_n_halfpoints,
                        &strategy_oudler);
  with_call = !!with_call;
  assert (with_call == 0 || with_call == 1);
  partner_known = !!partner_known;
  assert (partner_known == 0 || partner_known == 1);
  bid_has_petit = !!bid_has_petit;
  assert (bid_has_petit == 0 || bid_has_petit == 1);
  bid_has_excuse = !!bid_has_excuse;
  assert (bid_has_excuse == 0 || bid_has_excuse == 1);
  bid_has_twentyone = !!bid_has_twentyone;
  assert (bid_has_twentyone == 0 || bid_has_twentyone == 1);
  team_attack = !!team_attack;
  assert (team_attack == 0 || team_attack == 1);
  team_defence = !!team_defence;
  assert (team_defence == 0 || team_defence == 1);
  play_called_suit = !!play_called_suit;
  assert (play_called_suit == 0 || play_called_suit == 1);
  ally_may_be_minor_master = !!ally_may_be_minor_master;
  assert (ally_may_be_minor_master == 0 || ally_may_be_minor_master == 1);
  ennemy_may_be_minor_master = !!ennemy_may_be_minor_master;
  assert (ennemy_may_be_minor_master == 0 || ennemy_may_be_minor_master == 1);
  ally_may_be_trump_master = !!ally_may_be_trump_master;
  assert (ally_may_be_trump_master == 0 || ally_may_be_trump_master == 1);
  ennemy_may_be_trump_master = !!ennemy_may_be_trump_master;
  assert (ennemy_may_be_trump_master == 0 || ennemy_may_be_trump_master == 1);
  strategy_outbid_v = !!strategy_outbid_v;
  assert (strategy_outbid_v == 0 || strategy_outbid_v == 1);
  strategy_call_shortest = !!strategy_call_shortest;
  assert (strategy_call_shortest == 0 || strategy_call_shortest == 1);
  strategy_call_second_shortest = !!strategy_call_second_shortest;
  assert (strategy_call_second_shortest == 0
          || strategy_call_second_shortest == 1);
  strategy_call_second_longest = !!strategy_call_second_longest;
  assert (strategy_call_second_longest == 0
          || strategy_call_second_longest == 1);
  strategy_call_longest = !!strategy_call_longest;
  assert (strategy_call_longest == 0 || strategy_call_longest == 1);
  strategy_call_myself_v = !!strategy_call_myself_v;
  assert (strategy_call_myself_v == 0 || strategy_call_myself_v == 1);
  strategy_call_marriage_v = !!strategy_call_marriage_v;
  assert (strategy_call_marriage_v == 0 || strategy_call_marriage_v == 1);
  strategy_lead_called_suit = !!strategy_lead_called_suit;
  assert (strategy_lead_called_suit == 0 || strategy_lead_called_suit == 1);
  strategy_lead_trumps = !!strategy_lead_trumps;
  assert (strategy_lead_trumps == 0 || strategy_lead_trumps == 1);
  strategy_play_excuse = !!strategy_play_excuse;
  assert (strategy_play_excuse == 0 || strategy_play_excuse == 1);
  strategy_trump_master_v = !!strategy_trump_master_v;
  assert (strategy_trump_master_v == 0 || strategy_trump_master_v == 1);
  strategy_trump_obvious_master = !!strategy_trump_obvious_master;
  assert (strategy_trump_obvious_master == 0
          || strategy_trump_obvious_master == 1);
  strategy_oudler = !!strategy_oudler;
  assert (strategy_oudler == 0 || strategy_oudler == 1);
#define strategy_bid strategy_bid_v
#define strategy_outbid strategy_outbid_v
#define strategy_call_myself strategy_call_myself_v
#define strategy_call_marriage strategy_call_marriage_v
#define strategy_trump_master strategy_trump_master_v
#define yes 1
#define no 0
#include "tree-1.h"
  tree_1 = score;
#include "tree-2.h"
  tree_2 = score;
#include "tree-3.h"
  tree_3 = score;
#include "tree-4.h"
  tree_4 = score;
#include "tree-5.h"
  tree_5 = score;
#include "linear-1.h"
  linear_1 = score;
#include "linear-2.h"
  linear_2 = score;
#include "linear-3.h"
  linear_3 = score;
#include "linear-4.h"
  linear_4 = score;
#include "linear-5.h"
  linear_5 = score;
#include "stacking.h"
#undef strategy_bid
#undef strategy_outbid
#undef strategy_call_myself
#undef strategy_call_marriage
#undef strategy_trump_master
#undef yes
#undef no
  (void) tree_1;
  (void) tree_2;
  (void) tree_3;
  (void) tree_4;
  (void) tree_5;
  (void) linear_1;
  (void) linear_2;
  (void) linear_3;
  (void) linear_4;
  (void) linear_5;
  return score;
}

static double
cnn_predictor_eval_one (const TarotGame * game, const TarotGameEvent * event)
{
  TarotGame copy;
  TarotPlayer next;
  int has_next = (game_get_next (game, &next) == TAROT_GAME_OK);
  double predictions[5];
  size_t n_predictions;
  assert (has_next);
  game_copy (&copy, game);
  if (game_add_event (&copy, event) != TAROT_GAME_OK)
    {
      assert (0);
    }
  n_predictions =
    cnn_predict (&copy, sizeof (predictions) / sizeof (predictions[0]),
                 predictions);
  assert (n_predictions <= sizeof (predictions) / sizeof (predictions[0]));
  assert (next < n_predictions);
  return predictions[next];
}

static double
predictor_eval_one (int cnn, const TarotGame * game,
                    const TarotGameEvent * event)
{
  if (cnn)
    {
      return cnn_predictor_eval_one (game, event);
    }
  else
    {
      return stacking_predictor_eval_one (game, event);
    }
}

static int
predictor_select_bid (int cnn, const TarotGame * game, TarotBid * bid,
                      double *best_score)
{
  TarotBid minimum, candidate;
  TarotGameEvent e;
  TarotPlayer player;
  e.n = 0;
  if (game_get_hint_bid (game, &player, &minimum) != TAROT_GAME_OK)
    {
      return 1;
    }
  event_set_bid (&e, TAROT_PASS);
  *bid = TAROT_PASS;
  *best_score = predictor_eval_one (cnn, game, &e);
  for (candidate = minimum; candidate <= TAROT_DOUBLE_KEEP; candidate++)
    {
      double candidate_score;
      event_set_bid (&e, candidate);
      candidate_score = predictor_eval_one (cnn, game, &e);
      if (candidate_score > *best_score)
        {
          *bid = candidate;
          *best_score = candidate_score;
        }
    }
  return 0;
}

static int
predictor_select_decl (int cnn, const TarotGame * game, int *decl,
                       double *best_score)
{
  int allowed;
  TarotPlayer player;
  if (game_get_hint_decl (game, &player, &allowed) != TAROT_GAME_OK)
    {
      return 1;
    }
  *decl = 0;
  *best_score = 0;
  (void) cnn;
  return 0;
}

static int
predictor_select_call (int cnn, const TarotGame * game, TarotCard * call,
                       double *best_score)
{
  static const TarotSuit suits[4] = {
    TAROT_HEARTS,
    TAROT_CLUBS,
    TAROT_DIAMONDS,
    TAROT_SPADES
  };
  TarotNumber minimum, candidate;
  TarotGameEvent e;
  TarotPlayer player;
  int ret = 1;
  int isuit;
  e.n = 0;
  *best_score = -DBL_MAX;
  if (game_get_hint_call (game, &player, &minimum) != TAROT_GAME_OK)
    {
      return 1;
    }
  for (isuit = 0; isuit < 4; isuit++)
    {
      TarotSuit suit = suits[isuit];
      for (candidate = minimum; candidate <= TAROT_KING; candidate++)
        {
          TarotCard c;
          double candidate_score;
          if (tarot_of (candidate, suit, &c) != 0)
            {
              assert (0);
            }
          event_set_call (&e, c);
          candidate_score = predictor_eval_one (cnn, game, &e);
          if (candidate_score > *best_score)
            {
              *call = c;
              *best_score = candidate_score;
              ret = 0;
            }
        }
    }
  return ret;
}

static int
predictor_discard_try_changing_card (int cnn, const TarotGame * game,
                                     size_t n_candidates,
                                     const TarotCard * candidates,
                                     size_t n_discard,
                                     const TarotCard * discard_base,
                                     size_t i_card_changed,
                                     TarotCard * best_discard,
                                     double *best_score)
{
  TarotCard changed[6];
  size_t i;
  int has_changed = 0;
  TarotGameEvent e;
  TarotCard e_data[6];
  e.n = 6;
  e.data = e_data;
  assert (n_discard <= 6);
  assert (i_card_changed < 6);
  for (i = 0; i < n_discard; i++)
    {
      changed[i] = discard_base[i];
    }
  for (i = 0; i < n_candidates; i++)
    {
      changed[i_card_changed] = candidates[i];
      event_set_discard (&e, n_discard, changed);
      if (game_check_event (game, &e))
        {
          double this_score = predictor_eval_one (cnn, game, &e);
          if (this_score > *best_score)
            {
              size_t j;
              for (j = 0; j < n_discard; j++)
                {
                  best_discard[j] = changed[j];
                }
              *best_score = this_score;
              has_changed = 1;
            }
        }
    }
  return has_changed;
}

/* FIXME: there is a huge bias there, the discard will favor hearts */
static int
predictor_set_discard_greedy (int cnn, const TarotGame * game,
                              size_t n_candidates,
                              const TarotCard * candidates, size_t n_discard,
                              TarotCard * discard, double *best_score)
{
  size_t i;
  TarotGameEvent e;
  TarotCard e_data[6];
  TarotCard discard_candidate[6];
  int discard_changed;
  size_t i_card_to_change = 0;
  e.n = 6;
  e.data = e_data;
  assert (n_discard <= n_candidates);
  assert (n_discard <= 6);
  for (i = 0; i < n_discard; i++)
    {
      discard_candidate[i] = candidates[i];
    }
  event_set_discard (&e, n_discard, discard_candidate);
  /* Since the prio cards come before the additional in candidates,
   * then taking only the first cards is a valid discard. */
  assert (game_check_event (game, &e));
  for (i = 0; i < n_discard; i++)
    {
      discard[i] = discard_candidate[i];
    }
  *best_score = predictor_eval_one (cnn, game, &e);
  do
    {
      discard_changed =
        predictor_discard_try_changing_card (cnn, game, n_candidates,
                                             candidates, n_discard,
                                             discard_candidate,
                                             (i_card_to_change++) % n_discard,
                                             discard, best_score);
    }
  while (discard_changed);
  return 0;
}

static int
predictor_select_discard (int cnn, const TarotGame * game, size_t *n_cards,
                          TarotCard * discard, double *best_score)
{
  size_t n_discardable = 0;
  TarotCard discardable[78];
  TarotPlayer next_player;
  size_t discard_size, n_prio, n_additional, i;
  TarotCard prio[78];
  TarotCard additional[78];
  TarotCard actual_discard[6];
  int ret;
  if (game_get_hint_discard
      (game, &next_player, &discard_size, &n_prio, 0, 78, prio,
       &n_additional, 0, 78, additional) != TAROT_GAME_OK)
    {
      return 1;
    }
  /* discard_size is initialized, otherwise game_get_hint_discard
   * would have returned TAROT_GAME_NA */
  for (i = 0; i < n_prio; i++)
    {
      assert (n_discardable < 78);
      discardable[n_discardable++] = prio[i];
    }
  for (i = 0; i < n_additional; i++)
    {
      assert (n_discardable < 78);
      discardable[n_discardable++] = additional[i];
    }
  *n_cards = discard_size;
  *best_score = -DBL_MAX;
  assert (discard_size <= 6);
  ret =
    predictor_set_discard_greedy (cnn, game, n_discardable, discardable,
                                  discard_size, actual_discard, best_score);
  if (ret == 0)
    {
      for (i = 0; i < discard_size; i++)
        {
          discard[i] = actual_discard[i];
        }
    }
  return ret;
}

static int
predictor_select_card (int cnn, const TarotGame * game, TarotCard * card,
                       double *best_score)
{
  size_t n_candidates;
  TarotCard playable[78];
  TarotPlayer next_player;
  TarotGameEvent e;
  size_t i;
  e.n = 0;
  if (game_get_hint_card (game, &next_player, &n_candidates, 0, 78, playable)
      != TAROT_GAME_OK)
    {
      return 1;
    }
  assert (n_candidates <= 78);
  assert (n_candidates >= 1);
  event_set_card (&e, playable[0]);
  *best_score = predictor_eval_one (cnn, game, &e);
  *card = playable[0];
  for (i = 1; i < n_candidates; i++)
    {
      double candidate_score;
      event_set_card (&e, playable[i]);
      candidate_score = predictor_eval_one (cnn, game, &e);
      if (candidate_score > *best_score)
        {
          *card = playable[i];
          *best_score = candidate_score;
        }
    }
  return 0;
}

static void
predictor_analyze (TarotPredictor * predictor)
{
  TarotBid bid;
  int decl;
  TarotCard call;
  TarotCard discard[6];
  TarotCard card;
  size_t n_discard;
  if (predictor_select_bid (predictor->cnn, &(predictor->game),
                            &bid, &(predictor->best_score)) == 0)
    {
      event_set_bid (&(predictor->best_event), bid);
      predictor->can_play = 1;
    }
  else if (predictor_select_decl (predictor->cnn, &(predictor->game),
                                  &decl, &(predictor->best_score)) == 0)
    {
      event_set_decl (&(predictor->best_event), decl);
      predictor->can_play = 1;
    }
  else if (predictor_select_call (predictor->cnn, &(predictor->game),
                                  &call, &(predictor->best_score)) == 0)
    {
      event_set_call (&(predictor->best_event), call);
      predictor->can_play = 1;
    }
  else if ((predictor_select_discard (predictor->cnn,
                                      &(predictor->game),
                                      &n_discard,
                                      discard,
                                      &(predictor->best_score))) == 0)
    {
      assert (n_discard <= 6);
      event_set_discard (&(predictor->best_event), n_discard, discard);
      predictor->can_play = 1;
    }
  else if (predictor_select_card (predictor->cnn, &(predictor->game),
                                  &card, &(predictor->best_score)) == 0)
    {
      event_set_card (&(predictor->best_event), card);
      predictor->can_play = 1;
    }
}

static inline void
predictor_set_game (TarotPredictor * predictor, const TarotGame * game)
{
  TarotPlayer next_player;
  predictor->can_play = 0;
  game_initialize (&(predictor->game), 4, 0);
  if (game_get_next (game, &next_player) != TAROT_GAME_OK)
    {
      return;
    }
  if (game_copy_as (&(predictor->game), game, next_player) != TAROT_GAME_OK)
    {
      return;
    }
  predictor_analyze (predictor);
}

static inline const TarotGameEvent *
predictor_best (const TarotPredictor * predictor, double *score)
{
  if (predictor->can_play)
    {
      *score = predictor->best_score;
      return &(predictor->best_event);
    }
  else
    {
      return NULL;
    }
}

static inline size_t
predictor_eval (const TarotPredictor * predictor, size_t n_candidates,
                TarotGameEvent ** candidates, size_t start, size_t max,
                double *score)
{
  size_t i = 0;
  for (i = 0; i < n_candidates; i++)
    {
      if (!game_check_event (&(predictor->game), candidates[i]))
        {
          return i;
        }
      if (i >= start && i - start < max)
        {
          score[i] =
            predictor_eval_one (predictor->cnn, &(predictor->game),
                                candidates[i]);
        }
    }
  return n_candidates;
}

static size_t
column_type (TarotGameEventT t)
{
  switch (t)
    {
    case TAROT_DEAL_EVENT:
    case TAROT_DEAL_ALL_EVENT:
      return 0;
    case TAROT_BID_EVENT:
      return 1;
    case TAROT_DECL_EVENT:
      return 2;
    case TAROT_CALL_EVENT:
      return 3;
    case TAROT_DOG_EVENT:
      return 4;
    case TAROT_DISCARD_EVENT:
      return 5;
    case TAROT_HANDFUL_EVENT:
      return 6;
    case TAROT_CARD_EVENT:
      return 7;
    default:
      assert (0);
    }
  return 0;
}

static size_t
column_slam (void)
{
  return 8;
}

static size_t
column_bid (TarotBid b)
{
  switch (b)
    {
    case TAROT_PASS:
      return 9;
    case TAROT_TAKE:
      return 10;
    case TAROT_PUSH:
      return 11;
    case TAROT_STRAIGHT_KEEP:
      return 12;
    case TAROT_DOUBLE_KEEP:
      return 13;
    }
  assert (0);
}

static size_t
column_card (TarotCard c)
{
  return 14 + c;
}

static size_t
features_cnn_index (size_t page, size_t column, size_t row)
{
  return column + (row * 92) + (page * 6 * 92);
}

static void
features_cnn_addone (size_t n_players, size_t page, size_t column, size_t row,
                     double *data)
{
  size_t index = features_cnn_index (page, column, row);
  data[index] += 1.0;
  if (row + n_players < 6)
    {
      index = features_cnn_index (page, column, row + n_players);
      data[index] += 1.0;
    }
}

static void
features_set_type (size_t n_players, int page, TarotPlayer player,
                   TarotGameEventT t, double *data)
{
  features_cnn_addone (n_players, page, column_type (t), player, data);
}

static void
features_set_type_all (size_t n_players, int page, TarotGameEventT t,
                       double *data)
{
  size_t row;
  for (row = 0; row < n_players; row++)
    {
      features_cnn_addone (n_players, page, column_type (t), row, data);
    }
}

static void
features_set_slam (size_t n_players, int page, TarotPlayer player, int decl,
                   double *data)
{
  if (decl)
    {
      features_cnn_addone (n_players, page, column_slam (), player, data);
    }
}

static void
features_set_bid (size_t n_players, size_t page, TarotPlayer player,
                  TarotBid bid, double *data)
{
  features_cnn_addone (n_players, page, column_bid (bid), player, data);
}

static void
features_set_card (size_t n_players, size_t page, TarotPlayer player,
                   TarotCard card, double *data)
{
  features_cnn_addone (n_players, page, column_card (card), player, data);
}

static void
features_cnn_add_event (const TarotGame * base, const TarotGameEvent * event,
                        double *data)
{
  size_t n_players = game_n_players (base);
  int with_call = game_with_call (base);
  TarotPlayer player;
  int has_player = (game_get_next (base, &player) == TAROT_GAME_OK);
  TarotPlayer myself;
  TarotCard cards[78];
  size_t n_cards;
  const size_t max_cards = sizeof (cards) / sizeof (*cards);
  TarotPlayer owners[78];
  const size_t max_owners = sizeof (owners) / sizeof (*owners);
  size_t n_owners;
  TarotBid bid;
  int decl;
  TarotCard card;
  size_t i;
  TarotPlayer taker;
  int has_taker = (game_get_taker (base, &taker) == TAROT_GAME_OK);
  assert (with_call == 0 || with_call == 1);
  if (event_get_deal (event, &myself, &n_cards, 0, max_cards, cards)
      == TAROT_EVENT_OK)
    {
      assert (n_cards <= max_cards);
      features_set_type (n_players, 0, myself, TAROT_DEAL_EVENT, data);
      for (i = 0; i < n_cards; i++)
        {
          features_set_card (n_players, 0, myself, cards[i], data);
        }
    }
  if (event_get_deal_all (event, &n_owners, 0, max_owners, owners)
      == TAROT_EVENT_OK)
    {
      assert (n_owners <= max_owners);
      features_set_type_all (n_players, 0, TAROT_DEAL_ALL_EVENT, data);
      for (i = 0; i < n_owners; i++)
        {
          if (owners[i] < n_owners)
            {
              features_set_card (n_players, 0, owners[i], i, data);
            }
        }
    }
  if (event_get_bid (event, &bid) == TAROT_EVENT_OK)
    {
      assert (has_player);
      features_set_type (n_players, 1, player, TAROT_BID_EVENT, data);
      features_set_bid (n_players, 1, player, bid, data);
    }
  if (event_get_decl (event, &decl) == TAROT_EVENT_OK)
    {
      assert (has_player);
      features_set_type (n_players, 2, player, TAROT_BID_EVENT, data);
      features_set_slam (n_players, 2, player, decl, data);
    }
  if (event_get_call (event, &card) == TAROT_EVENT_OK)
    {
      assert (has_taker);
      features_set_type (n_players, 3, taker, TAROT_CALL_EVENT, data);
      features_set_card (n_players, 3, taker, card, data);
    }
  if (event_get_dog (event, &n_cards, 0, max_cards, cards) == TAROT_EVENT_OK)
    {
      assert (has_taker);
      assert (n_cards <= max_cards);
      features_set_type (n_players, 3 + with_call, taker, TAROT_DOG_EVENT,
                         data);
      for (i = 0; i < n_cards; i++)
        {
          features_set_card (n_players, 3 + with_call, taker, cards[i], data);
        }
    }
  if (event_get_discard (event, &n_cards, 0, max_cards, cards)
      == TAROT_EVENT_OK)
    {
      assert (has_taker);
      assert (n_cards <= max_cards);
      features_set_type (n_players, 4 + with_call, taker, TAROT_DISCARD_EVENT,
                         data);
      for (i = 0; i < n_cards; i++)
        {
          features_set_card (n_players, 4 + with_call, taker, cards[i], data);
        }
    }
  if (event_get_handful (event, &n_cards, 0, max_cards, cards)
      == TAROT_EVENT_OK)
    {
      assert (has_player);
      assert (n_cards <= max_cards);
      features_set_type (n_players, 4 + with_call, player,
                         TAROT_HANDFUL_EVENT, data);
      for (i = 0; i < n_cards; i++)
        {
          features_set_card (n_players, 4 + with_call, player, cards[i],
                             data);
        }
    }
  if (event_get_card (event, &card) == TAROT_EVENT_OK)
    {
      size_t i_trick;
      int has_trick =
        (game_get_current_trick (base, &i_trick) == TAROT_GAME_OK);
      assert (has_player);
      assert (has_trick);
      features_set_type (n_players, 5 + with_call + i_trick, player,
                         TAROT_CARD_EVENT, data);
      features_set_card (n_players, 5 + with_call + i_trick, player, card,
                         data);
    }
}

#define DIMENSION (92 * 6 * 32)

static size_t
features_cnn_add (const TarotGame * game, size_t max, double *data)
{
  TarotGame reconstruction;
  size_t n_players = game_n_players (game);
  int with_call = game_with_call (game);
  TarotGameIterator i;
  const TarotGameEvent *event;
  double my_data[DIMENSION] = { 0.0 };
  size_t my_dimension = DIMENSION;
  size_t ptr;
  game_initialize (&reconstruction, n_players, with_call);
  game_iterator_setup (&i, game);
  while ((event = tarot_game_iterator_next_value (&i)) != NULL)
    {
      features_cnn_add_event (&reconstruction, event, my_data);
      if (game_add_event (&reconstruction, event) != TAROT_GAME_OK)
        {
          assert (0);
        }
    }
  for (ptr = 0; ptr < my_dimension && ptr < max; ptr++)
    {
      data[ptr] = my_data[ptr];
    }
  return my_dimension;
}

#include <tarot/game_private_impl.h>
#include <tarot/game_event_private_impl.h>
#include <tarot/cards_private_impl.h>
#include <tarot/counter_private_impl.h>
#include <tarot/cnn_private_impl.h>

#endif /* H_TAROT_FEATURES_PRIVATE_IMPL_INCLUDED */
